@@ -3,16 +3,14 @@ package op
 import (
 	"errors"
 
-	"gopkg.in/square/go-jose.v2"
+	jose "github.com/go-jose/go-jose/v3"
 )
 
-var (
-	ErrSignerCreationFailed = errors.New("signer creation failed")
-)
+var ErrSignerCreationFailed = errors.New("signer creation failed")
 
 type SigningKey interface {
 	SignatureAlgorithm() jose.SignatureAlgorithm
-	Key() interface{}
+	Key() any
 	ID() string
 }
 
@@ -23,9 +21,9 @@ func SignerFromKey(key SigningKey) (jose.Signer, error) {
 			Key:   key.Key(),
 			KeyID: key.ID(),
 		},
-	}, &jose.SignerOptions{})
+	}, (&jose.SignerOptions{}).WithType("JWT"))
 	if err != nil {
-		return nil, ErrSignerCreationFailed //TODO: log / wrap error?
+		return nil, ErrSignerCreationFailed // TODO: log / wrap error?
 	}
 	return signer, nil
 }
@@ -34,5 +32,5 @@ type Key interface {
 	ID() string
 	Algorithm() jose.SignatureAlgorithm
 	Use() string
-	Key() interface{}
+	Key() any
 }
